// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SOUNDLNX.CC) Copyright 1998-1999 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com
//
// For info on this file please see sound.html in the docs directory
//
// This file implements the sound output to the Linux devices /dev/midi??
// for music and /dev/dsp for audio output.
//


#include "bochs.h"
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>


#if BX_USE_SB16_SMF
#define this NULL
#endif

struct bx_sb16_output_struct { 
  FILE *midi;
  char *wavedevice;
  int wave;
}; 

#define OUTPUTDATA ( (bx_sb16_output_struct*) (BX_SB16_THIS outputdata) ) 

int bx_sb16_c::initoutput()
{
  OUTPUTDATA = new bx_sb16_output_struct;

  if (OUTPUTDATA == NULL)
    {
	// allocation of struct failed, disable sound output
      writelog( BOTHLOG(2), "Couldn't allocate outputdata in soundwin.cc. Sound disabled.");
      bx_options.sb16.midimode = 0;
      bx_options.sb16.wavemode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  OUTPUTDATA->midi = NULL;
  OUTPUTDATA->wavedevice = NULL;
  OUTPUTDATA->wave = 0;

  // the output is always ready
  BX_SB16_THIS waveoutputready = 1;
  BX_SB16_THIS midioutputready = 1;

  return BX_SB16_OUTPUT_OK;
}


int bx_sb16_c::doneoutput()
{
  if (OUTPUTDATA != NULL)
    delete(OUTPUTDATA);

  OUTPUTDATA = NULL;

  return BX_SB16_OUTPUT_OK;
}


int bx_sb16_c::openmidioutput(char *device)
{
  if ( (device == NULL) || (strlen(device) < 1) )
    {
      bx_options.sb16.midimode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  OUTPUTDATA->midi = fopen(device,"w");

  if (OUTPUTDATA->midi == NULL)
    {
      writelog( MIDILOG(2), "Couldn't open midi output device %s.");
      bx_options.sb16.midimode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  return BX_SB16_OUTPUT_OK;
}


int bx_sb16_c::sendmidicommand(int delta, int command, int length, Bit8u data[])
{
  UNUSED(delta);
  //  bx_panic("Sendmidicommand!!");

  fputc(command, OUTPUTDATA->midi);
  fwrite(data, 1, length, OUTPUTDATA->midi);
  fflush(OUTPUTDATA->midi);       // to start playing immediately

  return BX_SB16_OUTPUT_OK;
}


int bx_sb16_c::closemidioutput()
{
  fclose(OUTPUTDATA->midi);

  return BX_SB16_OUTPUT_OK;
}


int bx_sb16_c::openwaveoutput(char *device)
{
  int length = strlen(device) + 1;

  if (OUTPUTDATA->wavedevice != NULL)
    delete(OUTPUTDATA->wavedevice);

  OUTPUTDATA->wavedevice = new char[length];

  if (OUTPUTDATA->wavedevice == NULL)
    {
      bx_options.sb16.wavemode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  strncpy(OUTPUTDATA->wavedevice, device, length);

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::startwaveplayback(int frequency, int bits, int stereo, int format)
{
  int fmt, ret, wave;
  int signeddata = format & 1;

  if ( (OUTPUTDATA->wavedevice == NULL) || (strlen(OUTPUTDATA->wavedevice) < 1) )
    {
      bx_options.sb16.wavemode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  wave = OUTPUTDATA->wave = open(OUTPUTDATA->wavedevice, O_WRONLY | O_NONBLOCK);

  if (OUTPUTDATA->wave == -1)
    {
      bx_options.sb16.wavemode = 0;
      return BX_SB16_OUTPUT_ERR;
    }

  // set frequency etc.
  ret = ioctl(wave = OUTPUTDATA->wave, SNDCTL_DSP_RESET);
  if (ret != 0)
    writelog( WAVELOG(4), "ioctl(SNDCTL_DSP_RESET): %s", 
	      strerror(errno));

  ret = ioctl(wave = OUTPUTDATA->wave, SNDCTL_DSP_SPEED, &frequency);
  if (ret != 0)
    writelog( WAVELOG(4), "ioctl(SNDCTL_DSP_SPEED, %d): %s", 
	      frequency, strerror(errno));

  ret = ioctl(wave = OUTPUTDATA->wave, SNDCTL_DSP_STEREO, &stereo);
  if (ret != 0)
    writelog( WAVELOG(4), "ioctl(SNDCTL_DSP_STEREO, %d): %s", 
	      stereo, strerror(errno));

  if (bits == 16)
    if (signeddata == 1)
      fmt = AFMT_S16_LE;
    else
      fmt = AFMT_U16_LE;
  else if (bits == 8)
    if (signeddata == 1)
      fmt = AFMT_S8;
    else
      fmt = AFMT_U8;
  else
    return BX_SB16_OUTPUT_ERR;

  ret = ioctl(OUTPUTDATA->wave, SNDCTL_DSP_SETFMT, &fmt);
  if (ret != 0)
    writelog( WAVELOG(4), "ioctl(SNDCTL_DSP_SETFMT, %d): %s",
	      fmt, strerror(errno));

  return BX_SB16_OUTPUT_OK;
}
				 
int bx_sb16_c::sendwavepacket(int length, Bit8u data[])
{
  write(OUTPUTDATA->wave, data, length);

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::stopwaveplayback()
{
  close(OUTPUTDATA->wave);

  return BX_SB16_OUTPUT_OK;
}

int bx_sb16_c::closewaveoutput()
{
  if (OUTPUTDATA->wavedevice != NULL)
    delete(OUTPUTDATA->wavedevice);

  OUTPUTDATA->wavedevice = NULL;

  return BX_SB16_OUTPUT_OK;
}
